import os
import sys
import unittest

import requests

from yubico_client import yubico
from yubico_client.otp import OTP
from yubico_client.py3 import PY3, unittest2_required
from yubico_client.yubico_exceptions import StatusCodeError
from yubico_client.yubico_exceptions import InvalidClientIdError
from yubico_client.yubico_exceptions import SignatureVerificationError
from yubico_client.yubico_exceptions import InvalidValidationResponse

if unittest2_required:
    import unittest2 as unittest  # NOQA
else:
    import unittest

LOCAL_SERVER = ('http://127.0.0.1:8881/wsapi/2.0/verify',)
LOCAL_SERVER_HTTPS = ('https://127.0.0.1:8881/wsapi/2.0/verify',)


class TestOTPClass(unittest.TestCase):
    def test_otp_class(self):
        otp1 = OTP('tlerefhcvijlngibueiiuhkeibbcbecehvjiklltnbbl')
        otp2 = OTP('jjjjjjjjnhe.ngcgjeiuujjjdtgihjuecyixinxunkhj',
                   translate_otp=True)

        self.assertEqual(otp1.device_id, 'tlerefhcvijl')
        self.assertEqual(otp2.otp,
                         'ccccccccljdeluiucdgffccchkugjcfditgbglbflvjc')

    def test_translation_multiple_interpretations(self):
        otp_str1 = 'vvbtbtndhtlfguefgluvbdcetnitidgkvfkbicevgcin'
        otp1 = OTP(otp_str1)
        self.assertEqual(otp1.otp, otp_str1)

    def test_translation_single_interpretation(self):
        otp_str1 = 'cccfgvgitchndibrrtuhdrgeufrdkrjfgutfjbnhhglv'
        otp_str2 = 'cccagvgitchndibrrtuhdrgeufrdkrjfgutfjbnhhglv'
        otp1 = OTP(otp_str1)
        otp2 = OTP(otp_str2)
        self.assertEqual(otp1.otp, otp_str1)
        self.assertEqual(otp2.otp, otp_str2)


class TestYubicoVerifySingle(unittest.TestCase):
    def setUp(self):
        yubico.DEFAULT_TIMEOUT = 2

        self.client_no_verify_sig = yubico.Yubico('1234', None,
                                                  api_urls=LOCAL_SERVER)
        self.client_verify_sig = yubico.Yubico('1234', 'secret123456',
                                               api_urls=LOCAL_SERVER)

    def test_invalid_custom_ca_certs_path(self):
        if hasattr(sys, 'pypy_version_info') or PY3:
            # TODO: Figure out why this breaks PyPy and 3.3
            return

        client = yubico.Yubico('1234', 'secret123456',
                               api_urls=LOCAL_SERVER_HTTPS,
                               ca_certs_bundle_path=os.path.abspath(__file__))

        try:
            client.verify('bad')
        except requests.exceptions.SSLError:
            pass
        else:
            self.fail('SSL exception was not thrown')

    def test_custom_ca_certs_path_invalid_path(self):
        expected_msg = ('Invalid value provided for ca_certs_bundle_path '
                        'argument')
        self.assertRaisesRegexp(ValueError, expected_msg,
                                yubico.Yubico, '1234', 'secret123456',
                                ca_certs_bundle_path='/does/not/exist.1')

    def test_custom_ca_certs_path(self):
        file_path = os.path.abspath(__file__)
        client = yubico.Yubico('1234', 'secret123456',
                               ca_certs_bundle_path=file_path)
        self.assertEqual(client._get_ca_bundle_path(), file_path)

    def test_replayed_otp(self):
        self._set_mock_action('REPLAYED_OTP')

        try:
            self.client_no_verify_sig.verify('bad')
        except StatusCodeError:
            e = sys.exc_info()[1]
            self.assertEqual(e.status_code, 'REPLAYED_OTP')

    def test_verify_bad_status_codes(self):
        for status in (set(yubico.BAD_STATUS_CODES) - set(['REPLAYED_OTP'])):
            self._set_mock_action(status)

            try:
                self.client_no_verify_sig.verify('bad')
            except Exception:
                e = sys.exc_info()[1]
                self.assertEqual(str(e), 'NO_VALID_ANSWERS')

    def test_verify_local_timeout(self):
        self._set_mock_action('timeout')

        try:
            self.client_no_verify_sig.verify('bad')
        except Exception:
            e = sys.exc_info()[1]
            self.assertEqual(str(e), 'NO_VALID_ANSWERS')

    def test_verify_invalid_signature(self):
        self._set_mock_action('no_signature_ok')

        try:
            self.client_verify_sig.verify('test')
        except SignatureVerificationError:
            pass
        else:
            self.fail('Exception was not thrown')

    def test_verify_no_such_client(self):
        self._set_mock_action('no_such_client')

        try:
            self.client_no_verify_sig.verify('test')
        except InvalidClientIdError:
            e = sys.exc_info()[1]
            self.assertEqual(e.client_id, '1234')
        else:
            self.fail('Exception was not thrown')

    def test_verify_ok_dont_check_signature(self):
        self._set_mock_action('no_signature_ok')

        status = self.client_no_verify_sig.verify('test')
        self.assertTrue(status)

    def test_verify_ok_check_signature(self):
        signature = \
            self.client_verify_sig.generate_message_signature('status=OK')
        self._set_mock_action('ok_signature', signature=signature)

        status = self.client_verify_sig.verify('test')
        self.assertTrue(status)

    def test_verify_invalid_otp_returned_in_the_response(self):
        self._set_mock_action('no_signature_ok_invalid_otp_in_response')

        try:
            self.client_no_verify_sig.verify('test')
        except InvalidValidationResponse:
            e = sys.exc_info()[1]
            self.assertTrue('Unexpected OTP in response' in e.message)
        else:
            self.fail('Exception was not thrown')

    def test_verify_invalid_nonce_returned_in_the_response(self):
        self._set_mock_action('no_signature_ok_invalid_nonce_in_response')

        try:
            self.client_no_verify_sig.verify('test')
        except InvalidValidationResponse:
            e = sys.exc_info()[1]
            self.assertTrue('Unexpected nonce in response' in e.message)
        else:
            self.fail('Exception was not thrown')

    def test_verify_multi_different_device_ids(self):
        otp_list = [
            'tlerefhcvijlngibueiiuhkeibbcbecehvjiklltnbbl',
            'blerefhcvijlngibueiiuhkeibbcbecehvjiklltnbbl',
        ]

        expected_msg = 'OTPs contain different device ids'

        self.assertRaisesRegexp(Exception, expected_msg,
                                self.client_no_verify_sig.verify_multi,
                                otp_list=otp_list)

    def test_verify_multi_single_otp(self):
        otp_list = [
            'tlerefhcvijlngibueiiuhkeibbcbecehvjiklltnbbl'
        ]

        expected_msg = 'otp_list needs to contain at least two OTPs'

        self.assertRaisesRegexp(ValueError, expected_msg,
                                self.client_no_verify_sig.verify_multi,
                                otp_list=otp_list)

    def test_verify_multi_too_much_time_passed_between_otp_generations(self):
        otp_list = [
            'tlerefhcvijlngibueiiuhkeibbcbecehvjiklltnbbl',
            'tlerefhcvijlngibueiiuhkeibbcbecehvjiklltnbbc',
        ]

        max_time_window = 7

        def mock_verify(*args, **kwargs):
            otp = args[0]

            response = {}

            if otp == 'tlerefhcvijlngibueiiuhkeibbcbecehvjiklltnbbl':
                timestamp = 1383997754 * 8
            elif otp == 'tlerefhcvijlngibueiiuhkeibbcbecehvjiklltnbbc':
                timestamp = (1383997754 + max_time_window + 1) * 8
            else:
                raise Exception('Invalid OTP')

            response['timestamp'] = timestamp
            return response

        self.client_no_verify_sig.verify = mock_verify

        expected_msg = ('More than 7 seconds have passed between generating '
                        'the first and the last OTP')

        self.assertRaisesRegexp(Exception, expected_msg,
                                self.client_no_verify_sig.verify_multi,
                                otp_list=otp_list,
                                max_time_window=max_time_window)

    def test_verify_multi_first_otp_is_older_than_last(self):
        otp_list = [
            'tlerefhcvijlngibueiiuhkeibbcbecehvjiklltnbbl',
            'tlerefhcvijlngibueiiuhkeibbcbecehvjiklltnbbc',
        ]

        max_time_window = 7

        def mock_verify(*args, **kwargs):
            otp = args[0]

            response = {}

            if otp == 'tlerefhcvijlngibueiiuhkeibbcbecehvjiklltnbbl':
                timestamp = (1383997754 + max_time_window + 1) * 8
            elif otp == 'tlerefhcvijlngibueiiuhkeibbcbecehvjiklltnbbc':
                timestamp = 1383997754 * 8
            else:
                raise Exception('Invalid OTP')

            response['timestamp'] = timestamp
            return response

        self.client_no_verify_sig.verify = mock_verify

        expected_msg = ('delta is smaller than zero. First OTP appears '
                        'to be older than the last one')

        self.assertRaisesRegexp(Exception, expected_msg,
                                self.client_no_verify_sig.verify_multi,
                                otp_list=otp_list,
                                max_time_window=max_time_window)

    def test_verify_multi_success(self):
        otp_list = [
            'tlerefhcvijlngibueiiuhkeibbcbecehvjiklltnbbl',
            'tlerefhcvijlngibueiiuhkeibbcbecehvjiklltnbbc',
        ]

        def mock_verify(*args, **kwargs):
            otp = args[0]

            response = {}

            if otp == 'tlerefhcvijlngibueiiuhkeibbcbecehvjiklltnbbl':
                timestamp = 1383997754 * 8
            elif otp == 'tlerefhcvijlngibueiiuhkeibbcbecehvjiklltnbbc':
                timestamp = (1383997754 + 2) * 8
            else:
                raise Exception('Invalid OTP')

            response['timestamp'] = timestamp
            return response

        self.client_no_verify_sig.verify = mock_verify

        status = self.client_no_verify_sig.verify_multi(otp_list=otp_list)
        self.assertTrue(status)

    def _set_mock_action(self, action, port=8881, signature=None):
        path = '/set_mock_action?action=%s' % (action)

        if signature:
            path += '&signature=%s' % (signature)

        requests.get(url='http://127.0.0.1:%s%s' % (port, path))


class TestAPIUrls(unittest.TestCase):
    def test_default_urls(self):
        client = yubico.Yubico('1234', 'secret123456')
        expected_urls = list(yubico.DEFAULT_API_URLS)

        for url in expected_urls:
            self.assertTrue(url.startswith('https://'))

        self.assertEqual(client.api_urls, expected_urls)

    def test_custom_urls(self):
        custom_urls = ['https://example.com/wsapi/2.0/verify',
                       'http://example.com/wsapi/2.0/verify',
                       'http://example.com/wsapi/2.0/verify']

        client = yubico.Yubico('1234', 'secret123456', api_urls=custom_urls)
        self.assertEqual(client.api_urls, custom_urls)

    def test_custom_urls_invalid_or_missing_scheme(self):
        invalid_urls = [
            '127.0.0.1:8000/test',
            'ftp.example.com/test',
            'example.com'
        ]

        for url in invalid_urls:
            expected_msg = r'URL ".+?" contains an invalid or missing scheme'
            self.assertRaisesRegexp(ValueError, expected_msg,
                                    yubico.Yubico, '1234', 'secret123456',
                                    api_urls=[url])

    def test_single_url(self):
        single_url = 'http://www.example.com'
        client = yubico.Yubico('1234', 'secret123456', api_urls=single_url)
        self.assertEqual(client.api_urls, [single_url])


if __name__ == '__main__':
    sys.exit(unittest.main())



from yubico_client.modhex import translate
from yubico_client.py3 import u


class OTP(object):
    """
    Class which holds data about an OTP.
    """

    def __init__(self, otp, translate_otp=True):
        """
        Represents an OTP token.

        :param otp: OTP token.
        :type: otp: ``str``

        :param translate_otp: True if the OTP should be translated.
        :type translate_otp: ``bool``
        """
        if translate_otp:
            self.otp = self.get_otp_modehex_interpretation(otp)
        else:
            self.otp = otp

        self.device_id = self.otp[:12]
        self.session_counter = None
        self.timestamp = None
        self.session_user = None

    def get_otp_modehex_interpretation(self, otp):
        """
        Return modhex interpretation of the provided OTP.

        If there are multiple interpretations available, first one is used,
        because if the OTP uses all 16 characters in its alphabet there is only
        one possible interpretation of that OTP.

        :return: Modhex interpretation of the OTP.
        :rtype: ``str``
        """
        try:
            interpretations = translate(u(otp))
        except Exception:
            return otp

        if len(interpretations) == 0:
            return otp
        elif len(interpretations) > 1:
            # If there are multiple interpretations first try to use the same
            # translation as the input OTP. If the one is not found, use the
            # random interpretation.
            if u(otp) in interpretations:
                return otp

        return interpretations.pop()

    def __repr__(self):
        return '%s, %s, %s' % (self.otp, self.device_id, self.timestamp)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Libcloud Python 2.x and 3.x compatibility layer
# Some methods bellow are taken from Django PYK3 port which is licensed under 3
# clause BSD license
# https://bitbucket.org/loewis/django-3k

# Taken from Apache Libcloud which is license under the Apache 2.0 License
import sys

PY3 = False
PY2 = False
PY25 = False
PY27 = False
PY3 = False
PY32 = False

if sys.version_info >= (2, 0) and sys.version_info < (3, 0):
    PY2 = True

if sys.version_info >= (2, 5) and sys.version_info <= (2, 6):
    PY25 = True

if sys.version_info >= (2, 7) and sys.version_info <= (2, 8):
    PY27 = True

if sys.version_info >= (3, 0):
    PY3 = True

if sys.version_info >= (3, 2) and sys.version_info < (3, 3):
    PY32 = True

if PY3:
    from urllib.parse import urlencode as urlencode
    from urllib.parse import unquote as unquote

    u = str

    def b(s):
        if isinstance(s, str):
            return s.encode('utf-8')
        elif isinstance(s, bytes):
            return s
        else:
            raise TypeError("Invalid argument %r for b()" % (s,))
else:
    from urllib import urlencode as urlencode  # NOQA
    from urllib import unquote as unquote  # NOQA

    u = unicode
    b = bytes = str

if PY27 or PY3:
    unittest2_required = False
else:
    unittest2_required = True



# -*- coding: utf-8 -*-
#
# Name: Yubico Python Client
# Description: Python class for verifying Yubico One Time Passwords (OTPs).
#
# Author: Tomaz Muraus (http://www.tomaz.me)
# License: BSD
#
# Copyright (c) 2010-2013, Tomaž Muraus
# Copyright (c) 2012, Yubico AB
# All rights reserved.

import re
import os
import sys
import time
import hmac
import base64
import hashlib
import threading
import logging

import requests

from yubico_client.otp import OTP
from yubico_client.yubico_exceptions import (StatusCodeError,
                                             InvalidClientIdError,
                                             InvalidValidationResponse,
                                             SignatureVerificationError)
from yubico_client.py3 import b
from yubico_client.py3 import urlencode
from yubico_client.py3 import unquote

logger = logging.getLogger('yubico.client')

# Path to the standard CA bundle file locations for most of the operating
# systems
COMMON_CA_LOCATIONS = [
    '/usr/local/lib/ssl/certs/ca-certificates.crt',
    '/usr/local/ssl/certs/ca-certificates.crt',
    '/usr/local/share/curl/curl-ca-bundle.crt',
    '/usr/local/etc/openssl/cert.pem',
    '/opt/local/lib/ssl/certs/ca-certificates.crt',
    '/opt/local/ssl/certs/ca-certificates.crt',
    '/opt/local/share/curl/curl-ca-bundle.crt',
    '/opt/local/etc/openssl/cert.pem',
    '/usr/lib/ssl/certs/ca-certificates.crt',
    '/usr/ssl/certs/ca-certificates.crt',
    '/usr/share/curl/curl-ca-bundle.crt',
    '/etc/ssl/certs/ca-certificates.crt',
    '/etc/pki/tls/cert.pem',
    '/etc/pki/CA/cacert.pem',
    'C:\Windows\curl-ca-bundle.crt',
    'C:\Windows\ca-bundle.crt',
    'C:\Windows\cacert.pem'
]

DEFAULT_API_URLS = ('https://api.yubico.com/wsapi/2.0/verify',
                    'https://api2.yubico.com/wsapi/2.0/verify',
                    'https://api3.yubico.com/wsapi/2.0/verify',
                    'https://api4.yubico.com/wsapi/2.0/verify',
                    'https://api5.yubico.com/wsapi/2.0/verify')

# How long to wait before the time out occurs
DEFAULT_TIMEOUT = 10

# How many seconds can pass between the first and last OTP generation so the
# OTP is still considered valid
DEFAULT_MAX_TIME_WINDOW = 5

BAD_STATUS_CODES = ['BAD_OTP', 'REPLAYED_OTP', 'BAD_SIGNATURE',
                    'MISSING_PARAMETER', 'OPERATION_NOT_ALLOWED',
                    'BACKEND_ERROR', 'NOT_ENOUGH_ANSWERS',
                    'REPLAYED_REQUEST']


class Yubico(object):
    def __init__(self, client_id, key=None, verify_cert=True,
                 translate_otp=True, api_urls=DEFAULT_API_URLS,
                 ca_certs_bundle_path=None):

        if ca_certs_bundle_path and \
           not self._is_valid_ca_bundle_file(ca_certs_bundle_path):
            raise ValueError(('Invalid value provided for ca_certs_bundle_path'
                             ' argument'))

        self.client_id = client_id

        if key is not None:
            key = base64.b64decode(key.encode('ascii'))

        self.key = key
        self.verify_cert = verify_cert
        self.translate_otp = translate_otp
        self.api_urls = self._init_request_urls(api_urls=api_urls)
        self.ca_certs_bundle_path = ca_certs_bundle_path

    def verify(self, otp, timestamp=False, sl=None, timeout=None,
               return_response=False):
        """
        Verify a provided OTP.

        :param otp: OTP to verify.
        :type otp: ``str``

        :param timestamp: True to include request timestamp and session counter
                          in the response. Defaults to False.
        :type timestamp: ``bool``

        :param sl: A value indicating percentage of syncing required by client.
        :type sl: ``int`` or ``str``

        :param timeout: Number of seconds to wait for sync responses.
        :type timeout: ``int``

        :param return_response: True to return a response object instead of the
                                status code. Defaults to False.
        :type return_response: ``bool``

        :return: True is the provided OTP is valid, False if the
        REPLAYED_OTP status value is returned or the response message signature
        verification failed and None for the rest of the status values.
        """
        ca_bundle_path = self._get_ca_bundle_path()

        otp = OTP(otp, self.translate_otp)
        rand_str = b(os.urandom(30))
        nonce = base64.b64encode(rand_str, b('xz'))[:25].decode('utf-8')
        query_string = self.generate_query_string(otp.otp, nonce, timestamp,
                                                  sl, timeout)

        threads = []
        timeout = timeout or DEFAULT_TIMEOUT
        for url in self.api_urls:
            thread = URLThread('%s?%s' % (url, query_string), timeout,
                               self.verify_cert, ca_bundle_path)
            thread.start()
            threads.append(thread)

        # Wait for a first positive or negative response
        start_time = time.time()
        while threads and (start_time + timeout) > time.time():
            for thread in threads:
                if not thread.is_alive():
                    if thread.exception:
                        raise thread.exception
                    elif thread.response:
                        status = self.verify_response(thread.response,
                                                      otp.otp, nonce,
                                                      return_response)

                        if status:
                            if return_response:
                                return status
                            else:
                                return True
                    threads.remove(thread)
            time.sleep(0.1)

        # Timeout or no valid response received
        raise Exception('NO_VALID_ANSWERS')

    def verify_multi(self, otp_list, max_time_window=DEFAULT_MAX_TIME_WINDOW,
                     sl=None, timeout=None):
        """
        Verify a provided list of OTPs.

        :param max_time_window: Maximum number of seconds which can pass
                                between the first and last OTP generation for
                                the OTP to still be considered valid.
        :type max_time_window: ``int``
        """

        # Create the OTP objects
        otps = []
        for otp in otp_list:
            otps.append(OTP(otp, self.translate_otp))

        if len(otp_list) < 2:
            raise ValueError('otp_list needs to contain at least two OTPs')

        device_ids = set()
        for otp in otps:
            device_ids.add(otp.device_id)

        # Check that all the OTPs contain same device id
        if len(device_ids) != 1:
            raise Exception('OTPs contain different device ids')

        # Now we verify the OTPs and save the server response for each OTP.
        # We need the server response, to retrieve the timestamp.
        # It's possible to retrieve this value locally, without querying the
        # server but in this case, user would need to provide his AES key.
        for otp in otps:
            response = self.verify(otp.otp, True, sl, timeout,
                                   return_response=True)

            if not response:
                return False

            otp.timestamp = int(response['timestamp'])

        count = len(otps)
        delta = otps[count - 1].timestamp - otps[0].timestamp

        # OTPs have an 8Hz timestamp counter so we need to divide it to get
        # seconds
        delta = delta / 8

        if delta < 0:
            raise Exception('delta is smaller than zero. First OTP appears to '
                            'be older than the last one')

        if delta > max_time_window:
            raise Exception(('More than %s seconds have passed between '
                            'generating the first and the last OTP.') %
                            (max_time_window))

        return True

    def verify_response(self, response, otp, nonce, return_response=False):
        """
        Returns True if the OTP is valid (status=OK) and return_response=False,
        otherwise (return_response = True) it returns the server response as a
        dictionary.

        Throws an exception if the OTP is replayed, the server response message
        verification failed or the client id is invalid, returns False
        otherwise.
        """
        try:
            status = re.search(r'status=([A-Z0-9_]+)', response) \
                       .groups()

            if len(status) > 1:
                message = 'More than one status= returned. Possible attack!'
                raise InvalidValidationResponse(message, response)

            status = status[0]
        except (AttributeError, IndexError):
            return False

        signature, parameters = \
            self.parse_parameters_from_response(response)

        # Secret key is specified, so we verify the response message
        # signature
        if self.key:
            generated_signature = \
                self.generate_message_signature(parameters)

            # Signature located in the response does not match the one we
            # have generated
            if signature != generated_signature:
                logger.warn("signature mismatch for parameters=%r", parameters)
                raise SignatureVerificationError(generated_signature,
                                                 signature)
        param_dict = self.get_parameters_as_dictionary(parameters)

        if 'otp' in param_dict and param_dict['otp'] != otp:
            message = 'Unexpected OTP in response. Possible attack!'
            raise InvalidValidationResponse(message, response, param_dict)

        if 'nonce' in param_dict and param_dict['nonce'] != nonce:
            message = 'Unexpected nonce in response. Possible attack!'
            raise InvalidValidationResponse(message, response, param_dict)

        if status == 'OK':
            if return_response:
                return param_dict
            else:
                return True
        elif status == 'NO_SUCH_CLIENT':
            raise InvalidClientIdError(self.client_id)
        elif status == 'REPLAYED_OTP':
            raise StatusCodeError(status)

        return False

    def generate_query_string(self, otp, nonce, timestamp=False, sl=None,
                              timeout=None):
        """
        Returns a query string which is sent to the validation servers.
        """
        data = [('id', self.client_id),
                ('otp', otp),
                ('nonce', nonce)]

        if timestamp:
            data.append(('timestamp', '1'))

        if sl is not None:
            if sl not in range(0, 101) and sl not in ['fast', 'secure']:
                raise Exception('sl parameter value must be between 0 and '
                                '100 or string "fast" or "secure"')

            data.append(('sl', sl))

        if timeout:
            data.append(('timeout', timeout))

        query_string = urlencode(data)

        if self.key:
            hmac_signature = self.generate_message_signature(query_string)
            hmac_signature = hmac_signature
            query_string += '&h=%s' % (hmac_signature.replace('+', '%2B'))

        return query_string

    def generate_message_signature(self, query_string):
        """
        Returns a HMAC-SHA-1 signature for the given query string.
        http://goo.gl/R4O0E
        """
        # split for sorting
        pairs = query_string.split('&')
        pairs = [pair.split('=', 1) for pair in pairs]
        pairs_sorted = sorted(pairs)
        pairs_string = '&' . join(['=' . join(pair) for pair in pairs_sorted])

        digest = hmac.new(self.key, b(pairs_string), hashlib.sha1).digest()
        signature = base64.b64encode(digest).decode('utf-8')

        return signature

    def parse_parameters_from_response(self, response):
        """
        Returns a response signature and query string generated from the
        server response. 'h' aka signature argument is stripped from the
        returned query string.
        """
        lines = response.splitlines()
        pairs = [line.strip().split('=', 1) for line in lines if '=' in line]
        pairs = sorted(pairs)
        signature = ([unquote(v) for k, v in pairs if k == 'h'] or [None])[0]
        # already quoted
        query_string = '&' . join([k + '=' + v for k, v in pairs if k != 'h'])

        return (signature, query_string)

    def get_parameters_as_dictionary(self, query_string):
        """ Returns query string parameters as a dictionary. """
        pairs = (x.split('=', 1) for x in query_string.split('&'))
        return dict((k, unquote(v)) for k, v in pairs)

    def _init_request_urls(self, api_urls):
        """
        Returns a list of the API URLs.
        """
        if not isinstance(api_urls, (str, list, tuple)):
            raise TypeError('api_urls needs to be string or iterable!')

        if isinstance(api_urls, str):
            api_urls = (api_urls,)

        api_urls = list(api_urls)

        for url in api_urls:
            if not url.startswith('http://') and \
               not url.startswith('https://'):
                raise ValueError(('URL "%s" contains an invalid or missing'
                                 ' scheme' % (url)))

        return list(api_urls)

    def _get_ca_bundle_path(self):
        """
        Return a path to the CA bundle which is used for verifying the hosts
        SSL certificate.
        """
        if self.ca_certs_bundle_path:
            # User provided a custom path
            return self.ca_certs_bundle_path

        # Return first bundle which is available
        for file_path in COMMON_CA_LOCATIONS:
            if self._is_valid_ca_bundle_file(file_path=file_path):
                return file_path

        return None

    def _is_valid_ca_bundle_file(self, file_path):
        return os.path.exists(file_path) and os.path.isfile(file_path)


class URLThread(threading.Thread):
    def __init__(self, url, timeout, verify_cert, ca_bundle_path=None):
        super(URLThread, self).__init__()
        self.url = url
        self.timeout = timeout
        self.verify_cert = verify_cert
        self.ca_bundle_path = ca_bundle_path
        self.exception = None
        self.request = None
        self.response = None

    def run(self):
        logger.debug('Sending HTTP request to %s (thread=%s)' % (self.url,
                                                                 self.name))
        verify = self.verify_cert

        if self.ca_bundle_path is not None:
            verify = self.ca_bundle_path
            logger.debug('Using custom CA bunde: %s' % (self.ca_bundle_path))

        try:
            self.request = requests.get(url=self.url, timeout=self.timeout,
                                        verify=verify)
            self.response = self.request.content.decode('utf-8')
        except requests.exceptions.SSLError:
            e = sys.exc_info()[1]
            self.exception = e
            self.response = None
        except Exception:
            e = sys.exc_info()[1]
            logger.error('Failed to retrieve response: ' + str(e))
            self.response = None

        args = (self.url, self.name, self.response)
        logger.debug('Received response from %s (thread=%s): %s' % args)



import sys

from yubico_client import Yubico
from yubico_client import yubico_exceptions
from yubico_client.py3 import PY3

if PY3:
    raw_input = input

client_id = raw_input('Enter your client id: ')
secret_key = raw_input('Enter your secret key (optional): ')
token = raw_input('Enter OTP token: ')

if not secret_key:
    secret_key = None

client = Yubico(client_id, secret_key)

try:
    status = client.verify(token)
except yubico_exceptions.InvalidClientIdError:
    e = sys.exc_info()[1]
    print('Client with id %s does not exist' % (e.client_id))
    sys.exit(1)
except yubico_exceptions.SignatureVerificationError:
    print('Signature verification failed')
    sys.exit(1)
except yubico_exceptions.StatusCodeError:
    e = sys.exc_info()[1]
    print('Negative status code was returned: %s' % (e.status_code))
    sys.exit(1)

if status:
    print('Success, the provided OTP is valid')
else:
    print('No response from the servers or received other negative '
          'status code')
